// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:28 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "plug.h"
#include "datablock.h"
#include "dagnode.h"
#include "twistlog.h"
#include <fstream>

// ************************************************************************* //
/// DESCRIPTION: Root of the 
// ************************************************************************* //
DagNode	*DagNode::_m_root = 0;

// ************************************************************************* //
/// DESCRIPTION: Returns the next free id number.
///   This avoids looking in all the file to find what's free.
// ************************************************************************* //
int	DagNode::freeTypeNbr()
{
  std::map<int, DagNode *(*)()>::iterator	cit;
  int	ret(0);

  for (cit = creators().begin(); cit != creators().end(); ++cit)
    if (ret < cit->first)
      ret = cit->first;
  return ret + 1;
}

// ************************************************************************* //
DagNode::DagNode(DagNode *parent):
  _m_parent(parent), _m_dataBlock(new DataBlock(*this))
{
  if (parent)
    DagNode::parent(*parent, *this);
}

// ************************************************************************* //
/// DESCRIPTION: Disconnects all data, and deletes children (children's
///  data are disconnected recursively, and children's children 
///  deleted as well). Removes this from parent.
// ************************************************************************* //
DagNode::~DagNode()
{
  disconnectAll(*this);
  if (_m_parent)
    _m_parent->remove(*this);
  while (!_m_children.empty())
    delete _m_children.front();
  delete _m_dataBlock;
}


// ************************************************************************* //
/// DESCRIPTION: Instance counters.
// ************************************************************************* //
std::map<int, int> &DagNode::counters()
{
  static std::map<int, int>	scounters;

  return scounters;
}

// ************************************************************************* //
/// DESCRIPTION: Creates instance name and number.
// ************************************************************************* //
void	DagNode::createInstName()
{
  _m_instNbr = counters()[getTypeNbr()];
  _m_instStr = id2names()[getTypeNbr()] + String(counters()[getTypeNbr()]++);
}

// ************************************************************************* //
DataBlock	&DagNode::getDataBlock() const
{
  return *_m_dataBlock;
}

// ************************************************************************* //
/// DESCRIPTION: Makes the 2nd node a child of the 1st node.
// ************************************************************************* //
void	DagNode::parent(DagNode	&parent,
			DagNode	&child)
{
  if (child._m_parent)
    unparent(*child._m_parent, child);
  parent._m_children.push_back(&child);
  child._m_parent = &parent;
}

// ************************************************************************* //
/// DESCRIPTION: Removes the relationship between the nodes.
// ************************************************************************* //
void	DagNode::unparent(DagNode	&parent,
			  DagNode	&child)
{
  std::list<DagNode*>::iterator	cit;

  for (cit = parent._m_children.begin(); cit != parent._m_children.end(); )
    if (*cit == &child)
      parent._m_children.erase(cit++);
    else
      ++cit;
  child._m_parent = 0;
}

// ************************************************************************* //
/// DESCRIPTION:
// ************************************************************************* //
void	DagNode::connect(Plug	&d1,
			 Plug	&d2)
{
  DataBlock::connect(d1, d2);
}

// ************************************************************************* //
/// DESCRIPTION: Inter-node connection. Overloads the above.
///  If connected attributes do not exists, program exits.
// ************************************************************************* //
void	DagNode::connect(DagNode	&node1,
			 const String	&sn1,
			 DagNode	&node2,
			 const String	&sn2)
{
  DataBlock	&db1(node1.getDataBlock());
  DataBlock	&db2(node2.getDataBlock());
  Plug	*d1(db1.getPlug(sn1));
  Plug	*d2(db2.getPlug(sn2));

  if (!d1)
    {
      std::cerr << "Error: " << sn1 << " is not a data attribute of node "
		<< node1.getTypeStr() << std::endl;
      std::exit(EXIT_FAILURE);
    }
  if (!d2)
    {
      std::cerr << "Error: " << sn2 << " is not a data attribute of node "
		<< node2.getTypeStr() << std::endl;
      std::exit(EXIT_FAILURE);
    }
  DataBlock::connect(*d1, *d2);
}

// ************************************************************************* //
/// DESCRIPTION: Disconnects the two plugs.
// ************************************************************************* //
void	DagNode::disconnect(DagNode		&node1,
			    const String	&sn1,
			    DagNode		&node2,
			    const String	&sn2)
{
  DataBlock	&db1(node1.getDataBlock());
  DataBlock	&db2(node2.getDataBlock());

  DataBlock::disconnect(*db1.getPlug(sn1), *db2.getPlug(sn2));
}

// ************************************************************************* //
/// DESCRIPTION: Disconnects the plug from any existing input or output.
// ************************************************************************* //
void	DagNode::disconnect(DagNode		&node,
			    const String	&sn)
{
  DataBlock	&db(node.getDataBlock());

  DataBlock::disconnect(*db.getPlug(sn));
}

// ************************************************************************* //
void	DagNode::disconnectAll(DagNode &node)
{
  DataBlock::disconnectAll(*node._m_dataBlock);
}

// ************************************************************************* //
double	DagNode::getAsDouble(const char *name) const
{
  return _m_dataBlock->getAsDouble(name);
}

// ************************************************************************* //
void	*DagNode::readAsCustom(const char *name) const
{
  return _m_dataBlock->readAsCustom(name);
}

// ************************************************************************* //
void	DagNode::readAsSignal(const char *name) const
{
  return _m_dataBlock->readAsSignal(name);
}

// ************************************************************************* //
bool	DagNode::readAsBool(const char *name) const
{
  return _m_dataBlock->readAsBool(name);
}

// ************************************************************************* //
int	DagNode::readAsInt(const char *name) const
{
  return _m_dataBlock->readAsInt(name);
}

// ************************************************************************* //
double	DagNode::readAsDouble(const char *name) const
{
  return _m_dataBlock->readAsDouble(name);
}

// ************************************************************************* //
const String	&DagNode::readAsString(const char *name) const
{
  return _m_dataBlock->readAsString(name);
}

// ************************************************************************* //
const Pnt2d	&DagNode::readAsPnt2d(const char *name) const
{
  return _m_dataBlock->readAsPnt2d(name);
}

// ************************************************************************* //
const Mat2i	&DagNode::readAsMat2i(const char *name) const
{
  return _m_dataBlock->readAsMat2i(name);
}

// ************************************************************************* //
const Pnt3d	&DagNode::readAsPnt3d(const char *name) const
{
  return _m_dataBlock->readAsPnt3d(name);
}

// ************************************************************************* //
const Vec3d	&DagNode::readAsVec3d(const char *name) const
{
  return _m_dataBlock->readAsVec3d(name);
}

// ************************************************************************* //
const Rgb3d	&DagNode::readAsRgb3d(const char *name) const
{
  return _m_dataBlock->readAsRgb3d(name);
}

// ************************************************************************* //
const Mat3d	&DagNode::readAsMat3d(const char *name) const
{
  return _m_dataBlock->readAsMat3d(name);
}

// ************************************************************************* //
const TwistLog	&DagNode::readAsTwistLog(const char *name) const
{
  return _m_dataBlock->readAsTwistLog(name);
}

// ************************************************************************* //
const Mat4d	&DagNode::readAsMat4d(const char *name) const
{
  return _m_dataBlock->readAsMat4d(name);
}

// ************************************************************************* //
const std::vector<int> &DagNode::readAsIntArray(const char *name) const
{
  return _m_dataBlock->readAsIntArray(name);
}

// ************************************************************************* //
const std::vector<double> &DagNode::readAsDoubleArray(const char *name) const
{
  return _m_dataBlock->readAsDoubleArray(name);
}

// ************************************************************************* //
const std::vector<Pnt2d> &DagNode::readAsPnt2dArray(const char *name) const
{
  return _m_dataBlock->readAsPnt2dArray(name);
}

// ************************************************************************* //
const Pnt2d &DagNode::readAsPnt2dArray(const char	*name,
				       const int	i) const
{
  return _m_dataBlock->readAsPnt2dArray(name, i);
}

// ************************************************************************* //
const std::vector<Pnt3d> &DagNode::readAsPnt3dArray(const char *name) const
{
  return _m_dataBlock->readAsPnt3dArray(name);
}

// ************************************************************************* //
const Pnt3d &DagNode::readAsPnt3dArray(const char	*name,
				       const int	i) const
{
  return _m_dataBlock->readAsPnt3dArray(name, i);
}

// ************************************************************************* //
const std::vector<Vec3d> &DagNode::readAsVec3dArray(const char *name) const
{
  return _m_dataBlock->readAsVec3dArray(name);
}

// ************************************************************************* //
const Vec3d &DagNode::readAsVec3dArray(const char	*name,
				       const int	i) const
{
  return _m_dataBlock->readAsVec3dArray(name, i);
}

// ************************************************************************* //
const std::vector<TwistLog> &
DagNode::readAsTwistLogArray(const char *name) const
{
  return _m_dataBlock->readAsTwistLogArray(name);
}

// ************************************************************************* //
const TwistLog &DagNode::readAsTwistLogArray(const char	*name,
					     const int	i) const
{
  return _m_dataBlock->readAsTwistLogArray(name, i);
}

// ************************************************************************* //
const std::list<Pnt2d> &DagNode::readAsPnt2dList(const char *name) const
{
  return _m_dataBlock->readAsPnt2dList(name);
}

// ************************************************************************* //
void	DagNode::writeAsCustom(const char	*name,
			       void		*value)
{
  return _m_dataBlock->writeAsCustom(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsSignal(const char	*name)
{
  return _m_dataBlock->writeAsSignal(name);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsBool(const char	*name,
			     const bool	value)
{
  return _m_dataBlock->writeAsBool(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsInt(const char	*name,
			    const int	value)
{
  return _m_dataBlock->writeAsInt(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsDouble(const char		*name,
			       const double	value)
{
  return _m_dataBlock->writeAsDouble(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsString(const char	*name,
			       const String	&value)
{
  return _m_dataBlock->writeAsString(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsPnt2d(const char	*name,
			      const Pnt2d	&value)
{
  return _m_dataBlock->writeAsPnt2d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsMat2i(const char	*name,
			      const Mat2i	&value)
{
  return _m_dataBlock->writeAsMat2i(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsPnt3d(const char	*name,
			      const Pnt3d	&value)
{
  return _m_dataBlock->writeAsPnt3d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsVec3d(const char	*name,
			      const Vec3d	&value)
{
  return _m_dataBlock->writeAsVec3d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsRgb3d(const char	*name,
			      const Rgb3d	&value)
{
  return _m_dataBlock->writeAsRgb3d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsMat3d(const char	*name,
			      const Mat3d	&value)
{
  return _m_dataBlock->writeAsMat3d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsTwistLog(const char	*name,
				 const TwistLog	&value)
{
  return _m_dataBlock->writeAsTwistLog(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsMat4d(const char	*name,
			      const Mat4d	&value)
{
  return _m_dataBlock->writeAsMat4d(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsIntArray(const char		*name,
				 const std::vector<int>	&value)
{
  return _m_dataBlock->writeAsIntArray(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsDoubleArray(const char			*name,
				    const std::vector<double>	&value)
{
  return _m_dataBlock->writeAsDoubleArray(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsPnt2dArray(const char			*name,
				   const std::vector<Pnt2d>	&value)
{
  return _m_dataBlock->writeAsPnt2dArray(name, value);
}

// ************************************************************************* //
void	DagNode::writeAsPnt2dArray(const char	*name,
				   const int	i,
				   const Pnt2d	&value)
{
  return _m_dataBlock->writeAsPnt2dArray(name, i, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsPnt3dArray(const char			*name,
				   const std::vector<Pnt3d>	&value)
{
  return _m_dataBlock->writeAsPnt3dArray(name, value);
}

// ************************************************************************* //
void	DagNode::writeAsPnt3dArray(const char	*name,
				   const int	i,
				   const Pnt3d	&value)
{
  return _m_dataBlock->writeAsPnt3dArray(name, i, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsVec3dArray(const char			*name,
				   const std::vector<Vec3d>	&value)
{
  return _m_dataBlock->writeAsVec3dArray(name, value);
}

// ************************************************************************* //
void	DagNode::writeAsVec3dArray(const char	*name,
				   const int	i,
				   const Vec3d	&value)
{
  return _m_dataBlock->writeAsVec3dArray(name, i, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsTwistLogArray(const char			*name,
				      const std::vector<TwistLog>	&value)
{
  return _m_dataBlock->writeAsTwistLogArray(name, value);
}

// ************************************************************************* //
void	DagNode::writeAsTwistLogArray(const char	*name,
				   const int	i,
				   const TwistLog	&value)
{
  return _m_dataBlock->writeAsTwistLogArray(name, i, value);
}

// ************************************************************************* //
/// DESCRIPTION: Shortcut.
// ************************************************************************* //
void	DagNode::writeAsPnt2dList(const char			*name,
				  const std::list<Pnt2d>	&value)
{
  return _m_dataBlock->writeAsPnt2dList(name, value);
}

// ************************************************************************* //
/// DESCRIPTION: Returns handle of name name. If such name isn't found within
///  this node, searches in subgraph.
// ************************************************************************* //
Plug	*DagNode::getPlug(const String &name)
{
  Plug	*hnd;

  if ((hnd = _m_dataBlock->getPlug(name)))
    return hnd;
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Return handle of id
// ************************************************************************* //
Plug	*DagNode::getPlug(const int id)
{
  Plug	*hnd;

  if ((hnd = _m_dataBlock->getPlug(id)))
    return hnd;

  std::list<DagNode*>::iterator	nit;

  // HERE!!!
  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((hnd = (*nit)->getPlug(id)))
      return hnd;
  return 0;
}

// ************************************************************************* //
void	DagNode::getPlugByType(const int		type,
			       std::list<Plug *>	&hnd)
{
  _m_dataBlock->getPlugByType(type, hnd);//
}

// ************************************************************************* //
/// DESCRIPTION: Given the path, finds the node.
/// Path syntax: Node1:Node2:Node3:Node4 ...
// ************************************************************************* //
DagNode	*DagNode::getNodeByPathRec(const String &path)
{
  if (path.len() == 0)
    return this;

  const String	nodeChild(path.chopBack(':'));

  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getInstStr() ==  nodeChild)
      return (*nit)->getNodeByPathRec(path.chopFront(':'));
  return 0;
}

// ************************************************************************* //
DagNode	*DagNode::getNodeByPath(const String &path)
{
  const String	spath = path.chopFront(':');
  
  if (spath.empty() && path != ":")
    return 0;
  return DagNode::root().getNodeByPathRec(spath);  
}

// ************************************************************************* //
Plug	*DagNode::getPlugByPathRec(const String &path)
{
  if (path.len() == 0)
    {
      return 0;
    }

  String	nodeChild = path.chopBack(':');

  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getInstStr() ==  nodeChild)
      return (*nit)->getPlugByPathRec(path.chopFront(':'));

  nodeChild = path.chopBack('.');

  String	attrChild = path.chopFront('.');

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getInstStr() ==  nodeChild)
      return (*nit)->getPlug(attrChild);
  return 0;
}

// ************************************************************************* //
Plug	*DagNode::getPlugByPath(const String &path)
{
  const String	spath = path.chopFront(':');
  
  if (spath.empty() && path != ":")
    return 0;
  return DagNode::root().getPlugByPathRec(spath);  
}

// ************************************************************************* //
/// DESCRIPTION: Returns the child node of instance id numbder instNbr.
// ************************************************************************* //
DagNode	*DagNode::getNode(const int typeNbr,
			  const int instNbr)
{
  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getTypeNbr() ==  typeNbr && 
	(*nit)->getInstNbr() ==  instNbr)
      return (*nit);
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Browse the 
// ************************************************************************* //
void	DagNode::getNodesByDepth(int			depth,
				 std::list<DagNode *>	&nodes)
{
  std::list<DagNode*>::const_iterator	nit;

  nodes.push_back(this);
  if (0 <= --depth)
    for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
      (*nit)->getNodesByDepth(depth, nodes);
}

// ************************************************************************* //
/// DESCRIPTION: Browse the 
// ************************************************************************* //
void	DagNode::getNodesAtDepth(int			depth,
				 std::list<DagNode *>	&nodes)
{
  if (0 <= --depth)
    {
      std::list<DagNode*>::const_iterator	nit;

      for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
	(*nit)->getNodesByDepth(depth, nodes);
    }
  else
    nodes.push_back(this);
}

// ************************************************************************* //
/// DESCRIPTION: Looks in the children of this for a node of type typeStr.
/// TODO: Node repository, to speed up scene traversal.
// ************************************************************************* //
void	DagNode::getNodesByType(const String		&typeStr,
				std::list<DagNode *>	&nodes)
{
  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    {
      if ((*nit)->getTypeStr() == typeStr)
	nodes.push_back(*nit);
      (*nit)->getNodesByType(typeStr, nodes);
    }
}

// ************************************************************************* //
void	DagNode::getChildrenByType(const String		&typeStr,
				   std::list<DagNode *>	&nodes)
{
  std::list<DagNode*>::const_iterator	nit;

  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    if ((*nit)->getTypeStr() == typeStr)
      nodes.push_back(*nit);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the first parent of type s.
///  e.g. this is used to get the transform parent.
// ************************************************************************* //
DagNode	*DagNode::getNodeParentByType(const String &s)
{
  for (DagNode *parent(_m_parent); parent; parent = parent->_m_parent)
    if (parent->getTypeStr() == s)
      return parent;
  return 0;
}

// ************************************************************************* //
/// DESCRIPTION: Deselects everything in the node, and children.
// ************************************************************************* //
void	DagNode::deselect()
{
  std::list<DagNode*>::iterator	nit;
  Plug	*hnd;

  if ((hnd = getPlug("sel")))
    hnd->writeAsInt(0);
  _m_dataBlock->deselect();
  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    (*nit)->deselect();
}

// ************************************************************************* //
/// DESCRIPTION: Returns a list of the selected nodes and subnodes, sorted in
///  selection index.
///  and subnodes. Give a mask???
// ************************************************************************* //
void	DagNode::selectedRec(std::list<DagNode*> &nodes)
{
  std::list<DagNode*>::iterator	nit;
  Plug	*hnd;
  
  if ((hnd = getPlug("sel")) && hnd->readAsInt())
    nodes.push_back(this);
  for (nit = _m_children.begin(); nit != _m_children.end(); ++nit)
    (*nit)->selectedRec(nodes);
}

// ************************************************************************* //
/// DESCRIPTION: Returns a list of the selected nodes.
///  and subnodes.
// ************************************************************************* //
void	DagNode::getNodesSelected(std::list<DagNode*> &nodes)
{
  std::list<DagNode*>::iterator	nit;
  std::list<DagNode*>	tmpnodes;
  DagNode	**buckets;
  unsigned	maxindex(0);

  DagNode::root().selectedRec(tmpnodes);

  for (nit = tmpnodes.begin(); nit != tmpnodes.end(); ++nit)
    maxindex = max(unsigned((*nit)->readAsInt("sel")), maxindex);

  buckets = new DagNode*[maxindex + 1];
  memset(buckets, 0, (maxindex + 1) * sizeof(DagNode*));

  for (nit = tmpnodes.begin(); nit != tmpnodes.end(); ++nit)
    buckets[(*nit)->readAsInt("sel")] = *nit;
  for (unsigned i(1), count(0); i <= maxindex; ++i)
    if (buckets[i])
      {
	nodes.push_front(buckets[i]);
	buckets[i]->writeAsInt("sel", ++count);
      }
}

// ************************************************************************* //
/// DESCRIPTION: Outputs a text version of the node and everything.
// ************************************************************************* //
void	DagNode::save(std::ostream	&os) const
{
  std::list<DagNode*>::const_iterator	nit;

  os << getTypeStr() << ' '
     << int(this) << ' '
     << getInstStr() << ' ' 
     << getInstNbr() << std::endl;
  _m_dataBlock->save(os);
  if (_m_children.empty())
    os << "{" << std::endl << "}" << std::endl;
  else
    {
      os << "{" << std::endl;
      for (nit = _m_children.begin(); nit != _m_children.end(); )
	{
	  (*nit)->save(os);
	  if (++nit != _m_children.end())
	    os << "," << std::endl;
	}
      os << "}" << std::endl;
    }
  return;
}

// ************************************************************************* //
/// DESCRIPTION: Load the data.
/// Assume the node has just been created. So the type is known.
/// 2. get the adress and put in table
// ************************************************************************* //
void	DagNode::load(std::istream		&is,
		      std::map<int, void *>	&v2real,
		      std::map<int, std::list<void **> >	&ref2v,
		      std::list<Plug*>		&notSaved)
{
  while (!_m_children.empty())
    delete _m_children.front();
  _m_children.clear();

  String	skip, typeStr;

  is >> _m_instStr;
  is >> _m_instNbr;

  counters()[getTypeNbr()] = max(_m_instNbr + 1, counters()[getTypeNbr()]);

  _m_dataBlock->load(is, v2real, ref2v, notSaved);
  is >> skip;
  while (skip != "}")
    {
      int	adress;

      is >> typeStr;
      if (typeStr == "}")
	break;
      is >> adress;

      DagNode *child = (*DagNode::creators()[DagNode::names2id()[typeStr]])();

      v2real[adress] = child;
      add(*child);
      child->load(is, v2real, ref2v, notSaved);
      is >> skip;
    }
  // TODO: call initialize to assign ids!
}

// ************************************************************************* //
/// DESCRIPTION: Deletes everything in the node except datablock.
///  This should be called on Scene, so that everything will be 
///  gone (cameras, gizmo, etc...). Do an initialize afterward.
// ************************************************************************* //
void	DagNode::clear()
{
  disconnectAll(*this);
  if (_m_parent)
    _m_parent->remove(*this);
  getDataBlock().clear();
  while (!_m_children.empty())
    delete _m_children.front();
  _m_children.clear();
  counters().clear();
}

// ************************************************************************* //
void	DagNode::pushName() const
{
  glPushName(0);//Node
  glPushName(getTypeNbr());//Type
  glPushName(getInstNbr());//Instance
}

// ************************************************************************* //
void	DagNode::popName() const
{
  glPopName();//Instance
  glPopName();//Type
  glPopName();//Node
}

// ************************************************************************* //
void	DagNode::fixme()
{
  std::list<DagNode*>::iterator	cit;

  initialize();
  for (cit = _m_children.begin(); cit != _m_children.end(); ++cit)
    (*cit)->fixme();
}

// ************************************************************************* //
/// DESCRIPTION: Node constructor repositoty. 
// ************************************************************************* //
std::map<int, DagNode *(*)()> &DagNode::creators()
{
  static std::map<int, DagNode *(*)()>	screators;

  return screators;
}

// ************************************************************************* //
/// DESCRIPTION: Node constructor repositoty. 
// ************************************************************************* //
std::map<int, String> &DagNode::id2names()
{
  static std::map<int, String>	snames;

  return snames;
}

// ************************************************************************* //
/// DESCRIPTION: Node constructor repositoty. 
// ************************************************************************* //
std::map<String, int> &DagNode::names2id()
{
  static std::map<String, int>	snames;

  return snames;
}
