// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Fri Oct 28 14:33:33 EDT 2005
// ************************************************************************* //
#ifndef __DATACLOCK_HH__
#define __DATACLOCK_HH__

// ************************************************************************* //
extern "C"
{
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
}

#ifdef _WIN32
#include <windows.h>
#endif//_WIN32

// ************************************************************************* //
/// DESCRIPTION: This class provides methods to measure a program's 
///   performance.
// ************************************************************************* //
struct Clock
{
  typedef enum {RealTime, CPUTime} Mode;
private:
  double		_m_time;
#ifndef _WIN32
  struct timeval	_m_te;
#else//_WIN32
  DWORD				_m_te;
#endif//_WIN32
  clock_t		_m_c;
  bool			_m_mode;
public:
  Clock(const Mode mode = RealTime);
  double	tic();
  double	time() const;
};

// ************************************************************************* //
/// DESCRIPTION: Constructor.
///  RealTime	- creates a clock for the time experienced by the user.
///  CPUTime	- creates a clock for the CPU time used by the program..
/// Both are similar. If you run multiple applications, you should use CPUTime
/// to prevent any bias in the timing.
// ************************************************************************* //
inline Clock::Clock(const Mode mode): _m_time(0.0), _m_mode(mode)
{
  if (_m_mode == RealTime)
	{
#ifndef _WIN32
	  gettimeofday(&_m_te, 0);// use getTicCounts for windows (or similar)
#else//_WIN32
	  _m_te = GetTickCount();
#endif//_WIN32
	}
  else
    _m_c = clock();
}

// ************************************************************************* //
/// DESCRIPTION: Returns the time since the clock was built.
/// WARNING: time is stored as a double.
// ************************************************************************* //
inline double	Clock::time() const
{
  return _m_time;
}

// ************************************************************************* //
/// DESCRIPTION: Returns the number of seconds since last call.
// ************************************************************************* //
inline double  Clock::tic()
{
  double	ticval;

  if (_m_mode == RealTime)
    {
#ifndef _WIN32
      struct timeval        newte;

	  gettimeofday(&newte, 0);
      ticval = (newte.tv_sec - _m_te.tv_sec) +
		  (newte.tv_usec - _m_te.tv_usec) * 1.0e-6;
#else//_WIN32
      DWORD	newte;

	  newte = GetTickCount();
      ticval = (newte - _m_te) * 1.0e-6;
#endif//_WIN32
      _m_te = newte;
    }
  else
    {
      const clock_t	newc = clock();

      ticval = (newc - _m_c)  / double(CLOCKS_PER_SEC);
      _m_c = newc;
    }
  _m_time += ticval;
  return ticval;
}

// ************************************************************************* //
#endif//__DATACLOCK_HH__
