// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:29 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include "nodecurrentcontrol.h"
#include "nodevortexring.h"
#include "datablock.h"
#include "plug.h"
#include "bezier_simplex.h"
#include "command.h"
#include "gbb.h"

// ************************************************************************* //
int	NCurrentControl::typeNbr(64);
String	NCurrentControl::typeStr("CurrentControl");

// ************************************************************************* //
void	NCurrentControl::initialize()
{
  DataBlock	&db(getDataBlock());

  NShape::initialize();

  db.addMat4d(Mat4d::identity(), "matParent", false);//position
  db.addInt(0, MININT, MAXINT, "f", false);
  db.addInt(0, MININT, MAXINT, "fPrev", false);//framePrevious
  db.addSignal("update", false);
  db.addSignal("path", false);//link to path curve
  db.addPnt3d(Pnt3d(0.0), "p", false);
  db.addTwistLog(TwistLog(0.0), "rlog", false);
  db.addSignal("next", false);
  db.addSignal("prev", false);
  db.addSignal("signal", false);
  db.addDouble(1.0, 0.0, MAXDOUBLE, "presence");// Presence, in [0, 1]
  db.addDouble(1.0, 0.0, MAXDOUBLE, "paddling");// Presence, in [0, 1]

  db.affect("f", "update");
  db.affect("p", "rlog");

  buildDagNodes();
}

// ************************************************************************* //
void	NCurrentControl::buildDagNodes()
{
  DagNode *oldParent(_m_parent);
  DagNode *clock = DagNode::getNodeByPath(":Clock0");
  DagNode *cpath = Command::run(String("new -type Bezier3d -path ") + path());
  DagNode *tr = Command::run(String("new -type Transform -path :"));

  if (oldParent)
    DagNode::parent(*oldParent, *tr);
  DagNode::parent(*tr, *this);
  DagNode::connect(*oldParent, "renderIn", *tr, "renderIn");
  DagNode::connect(*tr, "renderIn", *this, "renderIn");
  DagNode::connect(*tr, "matChild", *this, "matParent");
  DagNode::connect(*tr, "matChild", *cpath, "matParent");
  //DagNode::connect(*_m_parent, "renderIn", *this, "renderIn");
  DagNode::connect(*cpath, "signal", *this, "path");
  DagNode::connect(*clock, "f", *this, "f");
  DagNode::connect(*this, "renderIn", *cpath, "renderIn");
}

// ************************************************************************* //
/// DESCRIPTION: C1 piecewise function, numerically very close to C2.
///  Also numerically almost symmetric.
/// Old kernel: 1.0 + lr * lr * lr * (lr * (15.0 - 6.0 * lr) - 10.0);
// ************************************************************************* //
double	rbfPath(const double	l2,
		const double	r2)
{
  if (r2 < l2)
    return 0.0;
  return sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
/// DESCRIPTION: Controls the low frequencies of the vortex rings.
//  TODO: Connect to other children
//        pick pseudo-random children
// ************************************************************************* //
void	NCurrentControl::update(Plug &plugOut)
{
  const int	fPrev = readAsInt("fPrev");
  const int	f = readAsInt("f");

  std::list<DagNode*>::iterator	nit;
  std::list<DagNode*>	nodes;
  Plug	*data;

  getNodesAtDepth(1, nodes);
  for (nit = nodes.begin(); nit != nodes.end(); ++nit)
    {
      if ((*nit)->getTypeStr() == "VortexRing")
	{
	  if ((f == 0 && f < fPrev) || 
	      (*nit)->readAsDouble("ttd") < (*nit)->readAsDouble("t"))
	    Command::run(String("delete -path ") + (*nit)->path());
	  else
	    {
	      (*nit)->readAsSignal("update");
	      DagNode	*path = getPlug("path")->getInputNode();
	      const Pnt3d	centerPrev((*nit)->readAsPnt3d("centerPrev"));
	      const Pnt3d	center((*nit)->readAsPnt3d("center"));
	      const Vec3d	dspltWanted = center - centerPrev;
	      double		l0((*nit)->readAsDouble("l"));
	      const double	ltot(path->readAsDouble("ltot"));

	      l0 += dspltWanted.norm();
	      if (ltot < l0)
		{
		  std::vector<DagNode*> onodes;
		  
		  getPlug("next")->getOutputNodes(onodes);
		  (*nit)->writeAsDouble("l", l0 - ltot);
		  if (onodes.size() == 0)
		    {
		      DagNode	*rs = DagNode::getNodeByPath(":VortexSet0");
		      DagNode::parent(*rs, **nit);
		    }
		  else
		    {
		      int i = int((double(rand() - 1) / RAND_MAX) *
				  onodes.size());
		      
		      DagNode::parent(*onodes[i], **nit);
		    }
		}
	      else
		{
		  std::list<DagNode *> nodes;
		  const Mat4d	&matParent(readAsMat4d("matParent"));

		  getChildrenByType("Bezier3d", nodes);
		  path->writeAsDouble("l", l0);

		  const Pnt3d	&pl0 = matParent * path->readAsPnt3d("pl");
		  const Vec3d	&tl0 = matParent * path->readAsVec3d("tl");
		  const Pnt3d	centerNew0(center - tl0 * ((center - pl0) *
							   tl0));
		  const Vec3d	dspltEffect0 = centerNew0 - centerPrev;
		  const double	c0(dspltWanted.norm() / dspltEffect0.norm());

		  double	l1((*nit)->readAsDouble("l"));

		  l1 += c0 * dspltWanted.norm();

		  path->writeAsDouble("l", l1);
		  const Pnt3d	&pl1 = matParent * path->readAsPnt3d("pl");
		  const Vec3d	&tl1 = matParent * path->readAsVec3d("tl");
		  const Pnt3d	centerNew1(center - tl1 * ((center - pl1) *
							   tl1));
		  const Vec3d	dspltEffect1 = centerNew1 - centerPrev;
		  const double	c1(dspltWanted.norm() / dspltEffect1.norm());
		  const double	presence(readAsDouble("presence"));
		  const double	paddling(readAsDouble("paddling"));
	      
		  (*nit)->writeAsDouble("l", l1);
		  (*nit)->writeAsPnt3d("center",
				       center + (centerNew1 - center) * 
				       presence);

		  Vec3d	ex((*nit)->readAsVec3d("ex"));
		  Vec3d	ey((*nit)->readAsVec3d("ey"));
		  Vec3d	ez((*nit)->readAsVec3d("ez"));
		  const Vec3d	&dir = (*nit)->readAsVec3d("dir");
		  const double	dir2(dir.norm2());
		  const double	alignment = (dir ^ tl1).norm2();

		  if (1.0e-8 < dir2 && 1.0e-8 < alignment)
		    {
		      const double	dir1 = sqrt(dir2);
		      TwistLog	rot(dir1 * presence  * acos(dir / dir1 *
							    tl1.unit()),
				    (dir ^ tl1).unit(), center);
		      Mat4d	mrot = rot.exp();

		      ex = (mrot * ex).unit();
		      ey = (mrot * ey).unit();
		      ez = (mrot * ez).unit();

		      (*nit)->writeAsVec3d("ex", ex);
		      (*nit)->writeAsVec3d("ey", ey);
		      (*nit)->writeAsVec3d("ez", ez);

		      // Modifies paddling direction and magnitude
		      // from dir to tl1
		      // The norm of "paddleAmount" describe how 
		      // much the paddling is effective

		      const Vec3d &paddle = (*nit)->readAsVec3d("paddle");

		      (*nit)->writeAsVec3d("paddle",
					   (dir ^ tl1).unit() * paddle.norm()
					   );
		      (*nit)->writeAsDouble("paddleAmount", paddling *
					    acos(dir.unit() * tl1) *
					    paddle.norm());
		    }

		  /// check out the dir vector
		}
	    }
	}
      else if ((data = (*nit)->getPlug("update")))
	data->readAsSignal();
    }
  writeAsInt("fPrev", f);
  plugOut.writeAsSignal();
}

// ************************************************************************* //
void	NCurrentControl::glRender(Plug &plugOut)
{
  if (readAsInt("renderIn") != Gbb::Rendering)
    {
      std::list<DagNode*>::iterator	nit;
      std::list<DagNode*>	nodes;
      const int	sel(readAsInt("sel"));
      Vec3d	color;
      Plug	*data;

      if (sel == 1)
	color = Vec3d(1.0, 0.0, 0.5);
      else if (sel)
	color = Vec3d(1.0, 0.0, 0.0);
      else
	color = Vec3d(0.0, 0.0, 1.0);
      color.glColor();
      pushName();
      getNodesAtDepth(1, nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	if ((data = (*nit)->getPlug("renderOut")))
	  data->readAsSignal();
      popName();
    }
  plugOut.writeAsSignal();
}

// ************************************************************************* //
/// DESCRIPTION: GBB compute method.
// ************************************************************************* //
void	NCurrentControl::compute(Plug &plugOut)
{
  if (plugOut.shortName() == "rlog")
    {
      const Pnt3d	&p(readAsPnt3d("p"));
      TwistLog		rlog(0.0);
      std::list<DagNode*>::iterator	nit;
      std::list<DagNode*>	nodes;

      getNodesAtDepth(1, nodes);
      for (nit = nodes.begin(); nit != nodes.end(); ++nit)
	if ((*nit)->getTypeNbr() == NVortexRing::typeNbr)
	  {
	    (*nit)->writeAsPnt3d("p", p);
	    rlog += (*nit)->readAsTwistLog("rlog");
	  }
      plugOut.writeAsTwistLog(rlog);
    }
  else if (plugOut.shortName() == "renderOut")
    glRender(plugOut);
  else if (plugOut.shortName() == "update")
    update(plugOut);
}
