// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <iostream>
#include <fstream>
#include "plugmesh.h"
#ifdef _WIN32
#include "plugmesh.h"
#endif//_WIN32

// ************************************************************************* //
int	PlugMesh::typeNbr(72);
String	PlugMesh::typeStr("PlugMesh");

// ************************************************************************* //
/// DESCRIPTION: Returns true if point p is inside mesh.
///  Uses vertical axis z = (0,0,1) to cound 
///  1. Test if point is inside mesh bbox.
///  2. Test if ray intersects an odd number of triangles
///     For each triangle:
///    2.1. Test if ray intersects triangle bbox.
///    2.2. Test if ray can intersect triangle.
///    2.3. Find ray parameter to plane.
///    2.3. Project configuration onto simple 2d case.
/// WARNING: doesnt work if the test ray intersects an edge (i.e. no 
///  priority rule).
// ************************************************************************* //
bool	Mesh::contains(Pnt3d p) const
{
  unsigned	count(0);

  if (!_m_bbox.contains(p))
    return false;

  for (std::vector<Pnt3d>::const_iterator vit = _m_vertices.begin();
       vit != _m_vertices.end(); ++vit)
    if (p.x == vit->x && p.y == vit->y)
      {
	p.eqJitter(1.0e-8);
	break;
      }
  for (std::vector<unsigned>::const_iterator iit = _m_triangles.begin();
       iit != _m_triangles.end(); )
    {
      const Pnt3d	&x0(_m_vertices[*(iit++)]);
      const Pnt3d	&x1(_m_vertices[*(iit++)]);
      const Pnt3d	&x2(_m_vertices[*(iit++)]);
      const BBox3d	bboxt(x0, x1, x2);

      if (bboxt.pmin.x <= p.x && p.x <= bboxt.pmax.x &&
	  bboxt.pmin.y <= p.y && p.y <= bboxt.pmax.y && p.z <= bboxt.pmax.z)
	{
	  const Vec3d	n((x1 - x0) ^ (x2 - x0));

	  if (n.z != 0.0)
	    {
	      const double t(((x0 - p) * n) / n.z);

	      if (0.0 <= t)
		{
		  const Pnt3d	q(p.x, p.y, p.z + t);
		  const Vec3d	s(bboxt.pmax - bboxt.pmin);
		  const unsigned	i((s.x <= s.y && s.x <= s.z)? 1:
					  ((s.y <= s.z)? 2: 0));
		  const unsigned	j = (i + 1) % 3;
		  const Pnt2d	p0(x0(i), x0(j));
		  const Pnt2d	p1(x1(i), x1(j));
		  const Pnt2d	p2(x2(i), x2(j));
		  const Pnt2d	r(q(i), q(j));

		  if ((((p1 - p0) ^ (r - p0)) <= 0.0 &&
		       ((p2 - p1) ^ (r - p1)) <= 0.0 &&
		       ((p0 - p2) ^ (r - p2)) <= 0.0) ||
		      (0.0 <= ((p1 - p0) ^ (r - p0)) &&
		       0.0 <= ((p2 - p1) ^ (r - p1)) &&
		       0.0 <= ((p0 - p2) ^ (r - p2))))
		    ++count;
		}
	    }
	}
    }
  return count % 2;;
}

// ************************************************************************* //
/// DESCRIPTION: Square distance between the point p and the closest point to
///  p on the surface or inside the shape. Thus if p is inside the shape, 
///  0 is returned.
///  Holds min and max bounds to the distance.
///  For each triangle, computes two cheap min and max bounds.
// ************************************************************************* //
double	Mesh::dist2(const Pnt3d &p) const
{
  if (contains(p))
    return 0.0;
  
  double	d2min(MAXDOUBLE), d2tmp;

  for (std::vector<unsigned>::const_iterator iit = _m_triangles.begin();
       iit != _m_triangles.end(); )
    {
      const Pnt3d	&x0(_m_vertices[*(iit++)]);
      const Pnt3d	&x1(_m_vertices[*(iit++)]);
      const Pnt3d	&x2(_m_vertices[*(iit++)]);
      const BBox3d	bboxt(x0, x1, x2);

      if (bboxt.dist2(p) < d2min &&
	  (d2tmp = dist2Triangle(x0, x1, x2, p)) < d2min)
	d2min = d2tmp;
    }  
  return d2min;
}

// ************************************************************************* //
double	rbfMesh(const double	l2,
		const double	r2)
{
  return (r2 < l2)? 0.0: sq(4.0 - 20.0 / (l2 / r2 + 4.0));
}

// ************************************************************************* //
double	Mesh::field(const Pnt3d		&p,
		    const double	safeZone,
		    const double	smoothZone) const
{
  BBox3d	bbox;

  bbox.eqExpand(_m_bbox, safeZone + smoothZone);
  if (!bbox.contains(p))
    return 0.0;

  const double	d2 = dist2(p);

  if (sq(safeZone + smoothZone) <= d2)
    return 0.0;
  if (d2 <= sq(safeZone))
    return 1.0;
  return rbfMesh(sq(sqrt(d2) - safeZone), sq(smoothZone));
}

// ************************************************************************* //
/// DESCRIPTION: Squared shortest distance to triangle. By Mgc D. Eberly.
// ************************************************************************* //
double	dist2Triangle(const Pnt3d	&p0,
		      const Pnt3d	&p1,
		      const Pnt3d	&p2,
		      const Pnt3d	&p)
{
  const Pnt3d	&vb(p0);
  Vec3d		ve0, ve1, vd;

  ve0 = p1 - vb;
  ve1 = p2 - vb;
  vd = vb - p;

  const double	a(ve0 * ve0), b(ve0 * ve1), c(ve1 * ve1), d(ve0 * vd);
  const double	e(ve1 * vd), f(vd * vd);
  const double	det(a * c - b * b);
  double	s(b * e - c * d), t(b * d - a * e);

  if (s + t <= det)
    if (s < 0.0)
      if (t < 0.0) // region4
	if (d < 0.0)
	  if (-d >= a)
	    return a + 2.0 * d + f;
	  else
	    return f - d * d / a;
	else
	  if (e >= 0.0)
	    return f;
	  else if (-e >= c)
	    return c + 2.0 * e + f;
	  else
	    return f - e * e / c;
      else // region3
	if (e >= 0.0)
	  return f;
	else if (-e >= c)
	  return c + 2.0 * e + f;
	else
	  return f - e * e / c;
    else if (t < 0.0) // region5
      if (d >= 0.0)
	return f;
      else if (-d >= a)
	return a + 2.0 * d + f;
      else
	return f - d * d / a;
    else // region0
      {
	const double invDet(1.0 / det);

	s *= invDet, t *= invDet;
	return s * (a * s + b * t + 2.0 * d) +
	  t * (b * s + c * t + 2.0 * e) + f;
      }
  else
    {
      double tmp0, tmp1, numer, denom;

      if (s < 0.0)// region2
	if ((tmp1 = c + e) > (tmp0 = b + d))
	  if ((numer = tmp1 - tmp0) >= (denom = a - 2.0 * b + c))
	    return a + 2.0 * d + f;
	  else
	    {
	      s = numer / denom, t = 1.0 - s;
	      return s * (a * s + b * t + 2.0 * d) +
		t * (b * s + c * t + 2.0 * e) + f;
	    }
	else
	  if (tmp1 <= 0.0)
	    return c + 2.0 * e + f;
	  else if (e >= 0.0)
	    return f;
	  else
	    return f - e * e / c;
      else if (t < 0.0) // region6
	if ((tmp1 = a + d) > (tmp0 = b + e))
	  if ((numer = tmp1 - tmp0) >= (denom = a - 2.0 * b + c))
	    return c + 2.0 * e + f;
	  else
	    {
	      t = numer / denom, s = 1.0 - t;
	      return s * (a * s + b * t + 2.0 * d)
		+ t * (b * s + c * t + 2.0 * e) + f;
	    }
	else
	  if (tmp1 <= 0.0)
	    return a + 2.0 * d + f;
	  else if (d >= 0.0)
	    return f;
	  else
	    return f - d * d / a;
      else // region1
	if ((numer = c + e - b - d) <= 0.0)
	  return c + 2.0 * e + f;
	else
	  if (numer >= (denom = a - 2.0 * b + c))
	    return a + 2.0 * d + f;
	  else
	    {
	      s = numer / denom, t = 1.0 - s;
	      return s * (a * s + b * t + 2.0 * d) +
		t * (b * s + c * t + 2.0 * e) + f;
	    }
    }
}

// ************************************************************************* //
void	Mesh::glRender(const bool isFlatShaded) const
{
  if (_m_texName)
    {
      glEnable(GL_TEXTURE_2D);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glBindTexture(GL_TEXTURE_2D, _m_texName);
      // TODO: render texture!
      glBegin(GL_TRIANGLES);
      if (isFlatShaded)
	for (unsigned i = 0; i < _m_triangles.size(); )
	  {
	    const Pnt3d	&p0(_m_vertices[_m_triangles[i]]);
	    const Pnt2d	&uv0(_m_uvs[_m_texcoords[i]]);
	    ++i;
	    const Pnt3d	&p1(_m_vertices[_m_triangles[i]]);
	    const Pnt2d	&uv1(_m_uvs[_m_texcoords[i]]);
	    ++i;
	    const Pnt3d	&p2(_m_vertices[_m_triangles[i]]);
	    const Pnt2d	&uv2(_m_uvs[_m_texcoords[i]]);
	    ++i;

	    ((p1 - p0) ^ (p2 - p0)).glNormal();
	    uv0.glTexCoord();
	    p0.glVertex();
	    uv1.glTexCoord();
	    p1.glVertex();
	    uv2.glTexCoord();
	    p2.glVertex();
	  }
      else
	for (unsigned i = 0; i < _m_triangles.size(); )
	  {
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_uvs[_m_texcoords[i]].glTexCoord();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_uvs[_m_texcoords[i]].glTexCoord();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_uvs[_m_texcoords[i]].glTexCoord();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	  }
      glEnd();//GL_TRIANGLES
      glDisable(GL_TEXTURE_2D);
    }
  else
    {
      // TODO: render texture!
      glBegin(GL_TRIANGLES);
      if (isFlatShaded)
	for (unsigned i = 0; i < _m_triangles.size(); )
	  {
	    const Pnt3d	&p0(_m_vertices[_m_triangles[i++]]);
	    const Pnt3d	&p1(_m_vertices[_m_triangles[i++]]);
	    const Pnt3d	&p2(_m_vertices[_m_triangles[i++]]);

	    ((p1 - p0) ^ (p2 - p0)).glNormal();
	    p0.glVertex();
	    p1.glVertex();
	    p2.glVertex();
	  }
      else
	for (unsigned i = 0; i < _m_triangles.size(); )
	  {
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	    _m_normals[_m_triangles[i]].glNormal();
	    _m_vertices[_m_triangles[i]].glVertex();
	    ++i;
	  }
      glEnd();//GL_TRIANGLES
    }
}

// ************************************************************************* //
/// DESCRIPTION: Creates normal by averaging the faces normals.
// ************************************************************************* //
void	Mesh::createNormals()
{
  _m_normals.clear();
  _m_normals = std::vector<Vec3d>(_m_vertices.size());
  for (unsigned	i(0), n(_m_normals.size()); i < n; ++i)
    _m_normals[i] = Vec3d(0.0);
  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      const Pnt3d	&p0(_m_vertices[_m_triangles[i + 0]]);
      const Pnt3d	&p1(_m_vertices[_m_triangles[i + 1]]);
      const Pnt3d	&p2(_m_vertices[_m_triangles[i + 2]]);
      const Vec3d	normal(((p2 - p1) ^ (p0 - p1)).unit());
      
      _m_normals[_m_triangles[i + 0]] += normal;
      _m_normals[_m_triangles[i + 1]] += normal;
      _m_normals[_m_triangles[i + 2]] += normal;
    }
  for (unsigned	i(0), n(_m_normals.size()); i < n; ++i)
    _m_normals[i].eqUnit();
}

// ************************************************************************* //
/// DESCRIPTION: 
///  Given the vertices and triangles, creates the edge arrays.
// ************************************************************************* //
void	Mesh::createEdges()
{
  std::map<Mesh::Edge, unsigned>	edges;
  std::map<Mesh::Edge, unsigned>	tedges;
  std::map<Mesh::Edge, unsigned>::iterator	eit, ejt;

  _m_edges.clear();
  for (unsigned	i(0), n(_m_triangles.size()); i < n; i += 3)
    {
      unsigned	p0(_m_triangles[i + 0]);
      unsigned	p1(_m_triangles[i + 1]);
      unsigned	p2(_m_triangles[i + 2]);
      Mesh::Edge	e[3] = { Mesh::Edge(min(p0, p1), max(p0, p1)),
				 Mesh::Edge(min(p1, p2), max(p1, p2)),
				 Mesh::Edge(min(p2, p0), max(p2, p0)) };
      for (unsigned k(0); k < 3; ++k)
	if (!edges[e[k]])
	  edges[e[k]] = i + 1;
	else
	  tedges[e[k]] = i + 1;
    }
  _m_edges = std::vector<Mesh::Edge>(edges.size());
  _m_tedges = std::vector<Mesh::TEdge>(edges.size());
  eit = edges.begin();
  ejt = tedges.begin();
  for (unsigned i(0); i < edges.size(); ++eit, ++ejt, ++i)
    {
      _m_edges[i] = eit->first;
      _m_tedges[i] = Mesh::TEdge(eit->second - 1, ejt->second - 1);
    }
  _m_bbox = BBox3d(_m_vertices.front());
  for (std::vector<Pnt3d>::const_iterator vit = _m_vertices.begin();
       vit != _m_vertices.end(); ++vit)
    _m_bbox.fill(*vit);
}

// ************************************************************************* //
/// DESCRIPTION: Reads silly
// ************************************************************************* //
void	Mesh::readAsOBS(std::istream &is)
{
  unsigned	size;
  String	str, texFilename;

  // Vertices ************************************************************** //
  is >> str;// read vertices
  is >> str;// read {
  is >> size;
  _m_vertices = std::vector<Pnt3d>(size);
  for (std::vector<Pnt3d>::iterator vit = _m_vertices.begin();
       vit != _m_vertices.end(); ++vit)
    is >> *vit;
  is >> str;// read }

  // Normals *************************************************************** //
  is >> str;// read normals
  is >> str;// read {
  is >> size;
  _m_normals = std::vector<Vec3d>(size);
  for (std::vector<Vec3d>::iterator nit = _m_normals.begin();
       nit != _m_normals.end(); ++nit)
    is >> *nit;
  is >> str;// read }

  // Triangles ************************************************************* //
  is >> str;// read triangles
  is >> str;// read {
  is >> size;
  _m_triangles = std::vector<unsigned>(size);
  for (std::vector<unsigned>::iterator iit = _m_triangles.begin();
       iit != _m_triangles.end(); ++iit)
    is >> *iit;
  is >> str;// read }

  // Material ************************************************************** //
  is >> str;// read material or nomaterial
  if (str == "material")
    {
      is >> str;// read {
      is >> texFilename;

      QImage	qimage(texFilename.data());

      setTexture(qimage);

      is >> str;// read uvs
      is >> str;// read {
      is >> size;
      _m_uvs = std::vector<Pnt2d>(size);
      for (std::vector<Pnt2d>::iterator nit = _m_uvs.begin();
	   nit != _m_uvs.end(); ++nit)
	is >> *nit;
      is >> str;// read }

      is >> str;// read texcoords
      is >> str;// read {
      is >> size;
      _m_texcoords = std::vector<unsigned>(size);
      for (std::vector<unsigned>::iterator nit = _m_texcoords.begin();
	   nit != _m_texcoords.end(); ++nit)
	is >> *nit;
      is >> str;// read }

      is >> str;// read }
    }
  else
    {
      _m_texName = 0;
    }
  createEdges();
}

// ************************************************************************* //
/// DESCRIPTION:
///  newmesh -file cube.wrl
// ************************************************************************* //
void	Mesh::readAsWRL(std::istream &is)
{
  String	tmp;

  // Points **************************************************************** //
  is.seekg(0, std::ios::beg);
  do 
    {
      is >> tmp;
    }
  while (tmp != "point");

  is >> tmp; // [
  while (tmp != "]")
    {
      double px, py, pz;

      is >> px >> py >> pz >> tmp; // x y z , OR x y z ]
      _m_vertices.push_back(Pnt3d(px, py, pz));
    }
  // Normals *************************************************************** //
  std::cout << "creating normals" << std::endl;
  is.seekg(0, std::ios::beg);
  do 
    {
      is >> tmp;
    }
  while (tmp != "vector");
  is >> tmp; // [
  while (tmp != "]")
    {
      double nx, ny, nz;

      is >> nx >> ny >> nz >> tmp; // x y z , OR x y z ]
      _m_normals.push_back(Vec3d(nx, ny, nz));
    }
  //

  // Triangles ************************************************************* //
  std::cout << "creating triangles" << std::endl;
  is.seekg(0, std::ios::beg);
  do 
    {
      is >> tmp;
    }
  while (tmp != "coordIndex");
  is >> tmp; // [
  while (tmp != "]")
    {
      int i0, i1, i2, m1;

      is >> i0 >> i1 >> i2 >> m1 >> tmp;  // i i i -1 , OR i i i -1 ]
      _m_triangles.push_back(i0);
      _m_triangles.push_back(i1);
      _m_triangles.push_back(i2);
    }
  _m_texName = 0;
  std::cout << "creating edge" << std::endl;
  createEdges();
  createNormals();
}

// ************************************************************************* //
/// DESCRIPTION: Reads a word from the stream, separated by
/// '\n' or ' ', and '[', ']', '{', '}' and ',' are single letter words.
/// Note: if a word is longer than 256, will chop it in 256 character parts.
/// This function is messed up due to problems with MinGW
// ************************************************************************* //
std::istream	&operator>>(std::istream	&is,
							String			&s)
{
	char	buf[256];
	int		lbegin(0);
	char	c = is.get();

	for ( ; !is.eof() && c && (c == ' ' || c == '\n' || c == '	'); )
		c = is.get();
	if (c == '[' || c == ']' || c == '{' || c == '}' || c == ',')
	{
		s.resize(2);
		s[0] = c;
		s[1] = 0;
		return is;
	}
	for ( ; !is.eof() && c && (c == ' ' || c == '\n' || c == '	'); )
		c = is.get();
	for (; !is.eof() && c && (c != ' ' && c != '\n' && c != '[' &&
		c != ']' && c != '{' && c != '}' &&
		c != ',' && c != '	'); c = is.get())
		buf[lbegin++] = c;
	if (c == '[' || c == ']' || c == ',' || c == '{' || c == '}')
		is.putback(c);
	s.resize(lbegin + 1);
	for (int i(0); i < lbegin; i++)
		s[i] = buf[i];
	s[lbegin] = 0;
	return is;
}

// ************************************************************************* //
/// DESCRIPTION: Tries to guess the right format.
// ************************************************************************* //
std::istream	&operator>>(std::istream	&is,
			    Mesh		&m)
{
  char s[256];

  is.getline(s, 256);

  String	str(s);

  if (str.contains("OBS") != -1)
    m.readAsOBS(is);
  else if (str.contains("VRML") != -1)
    m.readAsWRL(is);
  return is;
}
