// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#include <QApplication>
#include <QMainWindow>
#include <QGLWidget>
#include <QLayout>
#include <QFileDialog>
#include <QMessageBox>
#include <fstream>

// ************************************************************************* //
#include "qtMainWidget.h"
#include "qtGlWidget.h"
#include "qtRenderEditor.h"
#include "mystring.h"
#include "gbb.h"
#include "commandcustom.h"
#include "nodescene.h"

// ************************************************************************* //
/// Part of the GBB C++ API
// ************************************************************************* //
MainWindow::MainWindow(QWidget		*parent):
  QMainWindow(parent, Qt::Window), _m_isFrozen(false)
{
  NScene	*scene = new NScene();

  mainwindow() = this;
  setWindowTitle("");

  scene->initialize();
  scene->createInstName();

  CentralWidget	*centralwidget = new CentralWidget(this);

  _m_glwidget = centralwidget->glwidget();
  _m_clockEditor = centralwidget->clockEditor();
  _m_rendereditor = new RenderEditor(_m_glwidget->context());

  QMenu	*menu = menuBar()->addMenu(tr("&File"));

  menu->addAction("&New", this, SLOT(newScene()), Qt::CTRL + Qt::Key_N);
  menu->addAction("&Open", this, SLOT(openScene()), Qt::CTRL + Qt::Key_O);
  menu->addSeparator();
  menu->addAction("&Save As", this, SLOT(saveScene()), Qt::CTRL + Qt::Key_S);
  menu->addAction("&Export", this, SLOT(exportParticles()));
  menu->addAction("&Fix Scene", this, SLOT(fixScene()));
  menu->addSeparator();
  menu->addAction("&Exit", this, SLOT(quit()), Qt::CTRL + Qt::Key_Q);

  _m_view = menuBar()->addMenu(tr("&View"));
  _m_view->addAction("&Save", _m_glwidget, SLOT(saveView()));
  _m_view->addAction("&Load", _m_glwidget, SLOT(loadView()));
  _m_view->addSeparator();
  _m_view->addAction("&Reset", _m_glwidget, SLOT(resetView()),
		     Qt::CTRL + Qt::Key_V);

  _m_new = menuBar()->addMenu(tr("&New"));
  _m_new->addAction("&Smoke Generator", this, SLOT(newSmokeGenerator()));
  _m_new->addAction("&SmokeGPU Generator", this, SLOT(newSmokeGPUGenerator()));
  _m_new->addSeparator();
  _m_new->addAction("&Current Generator", this, SLOT(newCurrentGenerator()));
  _m_new->addAction("&Noise Generator", this, SLOT(newNoiseGenerator()));
  _m_new->addAction("&Turbulence", this, SLOT(newTurbulence()));
  _m_new->addAction("&Wind", this, SLOT(newWind()));
  _m_new->addSeparator();
  _m_new->addAction("&Current Control", this, SLOT(newCurrentControl()));
  _m_new->addAction("&Ring Checkpoint ", this, SLOT(newRingCheckpoint()));
  _m_new->addSeparator();
  _m_new->addAction("&Camera", this, SLOT(newAimCamera()));

  _m_edit = menuBar()->addMenu(tr("&Edit"));
  _m_edit->addAction("&Render", this, SLOT(render()), Qt::CTRL + Qt::Key_R);
  _m_edit->addSeparator();
  _m_edit->addAction("&Insert Key", this, SLOT(setKey()), Qt::Key_S);
  _m_edit->addAction("&Delete Key", this, SLOT(remKey()), Qt::Key_D);
  _m_edit->addSeparator();
  _m_edit->addAction("&Parent", this, SLOT(parent()));
  _m_edit->addAction("&Bind to Control", this, SLOT(bindToControl()));
  _m_edit->addAction("&Attach Controls", this, SLOT(attachControls()));
  _m_edit->addAction("&Sway CurrentGenerator", this,
		      SLOT(newControlGenerator()));
  _m_easyGizmo = _m_edit->addAction("&Easy Gizmo", this, SLOT(slotEasyGizmo()),
				    Qt::CTRL + Qt::Key_E);
  _m_easyGizmo->setCheckable(true);
  _m_easyGizmo->setShortcut(Qt::CTRL + Qt::Key_E);
  _m_edit->addSeparator();
  _m_edit->addAction("&Edit SmokeParticles", this, SLOT(editSmokeParticles()));
  _m_edit->addAction("&Edit SmokeGPUParticles", this,
		     SLOT(editSmokeGPUParticles()));
  _m_edit->addAction("&Edit VortexSet", this, SLOT(editVortexSet()));
  _m_edit->addSeparator();
  _m_edit->addAction("&Edit Sky", this, SLOT(editSky()));
  _m_edit->addAction("&Edit FrontLayer", this, SLOT(editFrontLayer()));
  _m_edit->addAction("&Edit BackLayer", this, SLOT(editBackLayer()));

  _m_shading = menuBar()->addMenu(tr("&Shading"));
  _m_shading->addAction("&Smooth", _m_glwidget, SLOT(setGouraud()),
			Qt::CTRL + Qt::Key_G);
  _m_shading->addAction("&Wires", _m_glwidget, SLOT(setWireframe()),
			Qt::CTRL + Qt::Key_W);

  QMenu	*obstacle = menuBar()->addMenu(tr("&Obstacle"));

  obstacle->addAction("&New Obstacle", this, SLOT(createObstacle()));
  obstacle->addAction("&Bake", this, SLOT(bakeObstacle()));

  setCentralWidget(centralwidget);
  _m_glwidget->setFocus();
}

// ************************************************************************* //
void	MainWindow::setWindowTitle(const QString &s)
{
  QMainWindow::setWindowTitle(QString("GraphicsBlueBox ") + GBBVERSION +
			      " " + s);
}

// ************************************************************************* //
void	MainWindow::newScene()
{
  if (QMessageBox::question(this,
			    tr("New Scene?"),
			    tr("Do you want to create a new scene?"),
			    tr("&Yes"), tr("&Cancel"),
			    QString(), 0, 1) == 0)
    Command::run("newscene");
}

// ************************************************************************* //
void	MainWindow::openScene()
{
  QString s = QFileDialog::getOpenFileName(this, "Open",
					   ".", "scene (*.gbb)");

  if (s.isEmpty())
    return;
  if (!s.contains(".gbb"))
    s = s + ".gbb";
  //std::cout << "load scene: " << s.toAscii().data() << std::endl;
  Command::run(String("load -file ") + s.toAscii().data());
  //std::cout << "Loaded scene" << std::endl;
}

// ************************************************************************* //
void	MainWindow::exportParticles()
{
  QString s = QFileDialog::getSaveFileName(this, "Save As", ".",
					   "scene (*.txt)");
  if (s.isEmpty())
    return;
  if (!s.contains(".txt"))
    s = s + ".txt";
  Command::run(String("export -file ") + s.toAscii().data());
}

// ************************************************************************* //
void	MainWindow::saveScene()
{
  QString s = QFileDialog::getSaveFileName(this, "Save As", ".",
					   "scene (*.gbb)");
  if (s.isEmpty())
    return;
  if (!s.contains(".gbb"))
    s = s + ".gbb";
  Command::run(String("save -file ") + s.toAscii().data());
}

// ************************************************************************* //
extern void	deleteFbo();

// ************************************************************************* //
void	MainWindow::quit()
{
  if (QMessageBox::question(this,
			    tr("Exit?"),
			    tr("Do you really want to exit?"),
			    tr("&Yes"), tr("&Cancel"),
			    QString(), 0, 1) == 0)
  {
	  deleteFbo();
	  qApp->quit();
  }
  setFocus();
  activateWindow();
}

// ************************************************************************* //
void	MainWindow::fixScene()
{
  DagNode::root().fixme();  
}

// ************************************************************************* //
void	MainWindow::render()
{
  if (_m_rendereditor)
    {
      _m_rendereditor->show();
      _m_rendereditor->raise();
      _m_rendereditor->setFocus();
      _m_rendereditor->activateWindow();
    }
}

// ************************************************************************* //
void	MainWindow::newGuide()
{
  Command::run("new -type Guide -path :");
}

// ************************************************************************* //
void	MainWindow::newCurrentGenerator()
{
  Command::run("new -type CurrentGenerator -path :");
}

// ************************************************************************* //
void	MainWindow::newSmokeGenerator()
{
  Command::run("new -type SmokeGenerator -path :");
}

// ************************************************************************* //
void	MainWindow::newSmokeGPUGenerator()
{
  Command::run("new -type SmokeGPUGenerator -path :");
}

// ************************************************************************* //
void	MainWindow::newCurrentControl()
{
  Command::run("new -type CurrentControl -path :VortexSet0");
}

// ************************************************************************* //
void	MainWindow::newRingCheckpoint()
{
  Command::run("new -type RingCheckpoint -path :VortexSet0");
}

// ************************************************************************* //
void	MainWindow::slotEasyGizmo()
{
  const bool	val(_m_easyGizmo->isChecked());
  DagNode	*gizmo = DagNode::getNodeByPath(":Gizmo0");

  _m_easyGizmo->setChecked(val);
  gizmo->writeAsBool("isEasy", val);
  _m_glwidget->updateGL();
}

// ************************************************************************* //
void	  FrontPanel::paintEvent(QPaintEvent *)//e
{
  QString	version = QString("version ") + GBBVERSION;
  QFontMetrics	fm(font());
  QRect		qs(fm.boundingRect(version));
  QPainter	p(this);

  p.drawImage(0, 0, _m_image);
  p.setPen(Qt::white);
  p.drawText(width() - 4 - qs.width(), height() - qs.height() - 4, version);
  p.end();
}
