// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// URL: http://www.cs.otago.ac.nz/postgrads/alexis
/// E-MAIL: alexis@cs.otago.ac.nz
/// STARTED: Sun Dec 21 10:18:20 NZDT 2003
// ************************************************************************* //
#ifndef	__SWIRLINGSWEEPER_HH__
#define	__SWIRLINGSWEEPER_HH__

// ************************************************************************* //
#include "gfxtypes.h"
#include "twistlog.h"

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: 2-dimensional table, for storing precomputed rotations
///  and centers of rotation.
// ************************************************************************* //
template <class T>
class TableDyn
{
  const int	_m_i, _m_j;
  T		*_m_m;
public:
  TableDyn(const int i, const int j): _m_i(i), _m_j(j), _m_m(new T[i * j]) {}
  virtual ~TableDyn() { delete[] _m_m; }
  T		&operator()(const int i, const int j)
  { return _m_m[i * _m_j + j]; }
  const T	&operator()(const int i, const int j) const 
  { return _m_m[i * _m_j + j]; }
};

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: SwirlingSweeper Space deformation.
///  _m_nbSpheres	- number of swirls on a single ring.
///			  multiply by steps to get total nb of swirls.
///  _m_mlog		- precomputed rotations.
///  _m_fields		- temporary stores the field values of one ring.
///  _m_sphereCenters	- precomputed swirl centers.
///  _m_k		- radius of a swirl.
///  _m_invk, _m_k2	- precomputed functions of _m_k.
///  _m_transf		- input translation.
///  _m_steps		- number of steps.
///  _m_size		- radius of ring.
///  _m_bbox		- bounding box.
/// 
/// USAGE: Call constructor with appropriate parameters, then sweep on data:
///
///   list<std::pair<Pnt3d, Vec3d> >	data;
///   SwirlingSweeper	sws(center, s, transf);
///
///   sws.sweep(data);
///
/// For detail on technique, see URL:
///   http://www.cs.otago.ac.nz/postgrads/alexis/publications.html
// ************************************************************************* //
class SwirlingSweeper
{
  const unsigned 	_m_nbSwirls;
  TableDyn<TwistLog>	*_m_mlog;
  double		*_m_fields;
  TableDyn<Pnt3d>	*_m_sphereCenters;
  double		_m_k, _m_invk, _m_k2;
  const Vec3d	_m_transf;
  mutable unsigned	_m_istep;
  unsigned		_m_steps;
  const double		_m_size;
  BBox3d		_m_bbox;
protected:
  void		sweepStepPoint(Pnt3d &s) const;
  void		sweepStepPair(Pnt3d &s) const;
  void		sweepStepPair(std::pair<Pnt3d, Vec3d> &s) const;
  double	field(const Pnt3d &p, const unsigned i) const;
public:
  SwirlingSweeper(const Pnt3d &c, const Pnt3d &q, const double s,
		  const int steps = -1);
  virtual ~SwirlingSweeper();
  void		sweep(std::list<std::pair<Pnt3d, Vec3d> > &l) const;
  void		sweep(std::list<Pnt3d> &l) const;
  void		sweep(Pnt3d &l) const;
  const BBox3d	&getBBox() const { return _m_bbox; }
};

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Value constructor.
///  c		- source
///  q		- destination
///  size		- radius of safety around center c.
///  steps	- use default value
// ************************************************************************* //
inline SwirlingSweeper::SwirlingSweeper(const Pnt3d &c,
					const Pnt3d &q,
					const double size,
					const int steps):
  _m_nbSwirls(8), _m_fields(new double[_m_nbSwirls]),
  _m_k(0.5 * size), _m_invk(1.0 / _m_k), _m_k2(_m_k * _m_k),
  _m_transf(q - c), _m_istep(0), _m_size(size / 4.0)
{
  // Precompute steps
  if (steps == -1)
    {
      const double	s(ceil(1.875 * 2.0 * _m_transf.norm() / _m_size));
      _m_steps = (s <= 1.0)? 1: unsigned(s);
    }
  else
    _m_steps = steps;

  // Precompute rotations
  const Vec3d	z(_m_transf.unit());
  Vec3d		x((fabs(z.x) < 0.5)? Vec3d(1.0, 0.0, 0.0):
			  Vec3d(0.0, 1.0, 0.0));

  x = (x - (x * z) * z).unit();

  const Vec3d	y(z ^ x);
  const double		astep(2.0 * M_PI / _m_nbSwirls);
  const double		angle((2.0 * _m_transf.norm() / 
			       (_m_steps * _m_size * _m_nbSwirls)));
  _m_mlog = new TableDyn<TwistLog>(_m_steps, _m_nbSwirls);
  _m_sphereCenters = new TableDyn<Pnt3d>(_m_steps, _m_nbSwirls);
  for (unsigned istep(0); istep < _m_steps; ++istep)
    {
      const Pnt3d	ci(c + _m_transf * (istep / double(_m_steps)));

      for (unsigned i(0), n(_m_nbSwirls); i < n; ++i)
	{
	  const double		cosa(cos(i * astep)), sina(sin(i * astep));
	  const Vec3d	nn(y * (cosa * angle) - x * (sina * angle));

	  (*_m_sphereCenters)(istep, i) = 
	    ci + x * (cosa * _m_size) + y * (sina * _m_size);
	  (*_m_mlog)(istep, i) = 
	    TwistLog(nn, Vec3d((*_m_sphereCenters)(istep, i)) ^ nn);
	}
    }
  // Precompute bbox.
  _m_bbox = BBox3d(c);
  _m_bbox.fill(c + _m_transf);
  _m_bbox.eqExpand(4.0 * _m_size);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Destructor.
// ************************************************************************* //
inline SwirlingSweeper::~SwirlingSweeper()
{
  delete _m_mlog;
  delete _m_sphereCenters;
  delete[] _m_fields;
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Returns the field of the spheres forming the tool.
// ************************************************************************* //
inline double	SwirlingSweeper::field(const Pnt3d	&p,
				       const unsigned	i) const
{
  const double	l2((p - (*_m_sphereCenters)(_m_istep, i)).norm2());

  if (_m_k2 <= l2)
    return 0.0;

  const double	l(sqrt(l2)), dk(l * _m_invk);

  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Blends the rotations. Uses quick exponential.
// ************************************************************************* //
inline void	SwirlingSweeper::sweepStepPoint(Pnt3d	&p) const
{
  TwistLog	m(0.0);
  double	sumf(0.0), f;
  
  for (unsigned i(0); i < _m_nbSwirls; ++i)
    if ((f = field(p, i)))
      _m_fields[i] = f, sumf += f;
    else
      _m_fields[i] = 0.0;
  if (sumf == 0.0)
    return;
  for (unsigned i(0); i < _m_nbSwirls; ++i)
    if ((f = _m_fields[i]) == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (f != 0.0)
      m += (*_m_mlog)(_m_istep, i) * f;
  m.exp(p);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Blends the rotations. Uses quick exponential.
// ************************************************************************* //
inline void	SwirlingSweeper::sweepStepPair(Pnt3d	&p) const
{
  TwistLog	m(0.0);

  for (unsigned i(0); i < _m_nbSwirls; ++i)
    if (_m_fields[i] == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (_m_fields[i] != 0.0)
      m += (*_m_mlog)(_m_istep, i) * field(p, i);
  m.exp(p);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Deforms a point and its normal.
/// NOTE: The matrix j is not the jacobian, but it is proportional to it,
///  hence adequate for computing the normal.
// ************************************************************************* //
inline void SwirlingSweeper::sweepStepPair(std::pair<Pnt3d, Vec3d> &s) const
{
  TwistLog	m(0.0);
  double	sumf(0.0), f;

  for (unsigned i(0); i < _m_nbSwirls; ++i)
    if ((f = field(s.first, i)))
      _m_fields[i] = f, sumf += f;
    else
      _m_fields[i] = 0.0;
  if (sumf == 0.0)
    return;
  for (unsigned i(0); i < _m_nbSwirls; ++i)
    if ((f = _m_fields[i]) == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (f != 0.0)
      m += (*_m_mlog)(_m_istep, i) * f;

  Pnt3d	&p(s.first), px(p), py(p), pz(p);

  m.exp(p);

  px.x += 1.0e-8, py.y += 1.0e-8, pz.z += 1.0e-8;
  sweepStepPair(px);
  sweepStepPair(py);
  sweepStepPair(pz);

  Mat3d	j(px - p, py - p, pz - p);

  j.eqCommatrix();
  s.second *= j;
  s.second.eqUnit();
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Deforms a list of points and their corresponding normals.
// ************************************************************************* //
inline void SwirlingSweeper::sweep(std::list<std::pair<Pnt3d,
				   Vec3d> > &l) const
{
  std::list<std::pair<Pnt3d, Vec3d> >::iterator	sit;

  if (!_m_steps)
    return;
  for (_m_istep = 0; _m_istep < _m_steps; ++_m_istep)
    for (sit = l.begin(); sit != l.end(); ++sit)
      if (_m_bbox.contains(sit->first))
	sweepStepPair(*sit);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Deforms a list of points. For the sake of 
///  efficiency, preferably use above method.
// ************************************************************************* //
inline void SwirlingSweeper::sweep(std::list<Pnt3d> &l) const
{
  std::list<Pnt3d>::iterator	sit;

  if (!_m_steps)
    return;
  for (_m_istep = 0; _m_istep < _m_steps; ++_m_istep)
    for (sit = l.begin(); sit != l.end(); ++sit)
      if (_m_bbox.contains(*sit))
	sweepStepPoint(*sit);
}

// ************************************************ AUTHOR: Alexis Angelidis //
/// DESCRIPTION: Deforms a list of points. For the sake of 
///  efficiency, preferably use above method.
// ************************************************************************* //
inline void SwirlingSweeper::sweep(Pnt3d &p) const
{
  if (!_m_steps)
    return;
  for (_m_istep = 0; _m_istep < _m_steps; ++_m_istep)
    if (_m_bbox.contains(p))
      sweepStepPoint(p);
}

// ************************************************ AUTHOR: Alexis Angelidis //
#endif//__SWIRLINGSWEEPER_HH__
