// ************************************************************************* //
/// GBB - Graphics Blue Box, by Alexis Angelidis
/// $Author: silex $
/// $Date: 2006/01/11 21:56:30 $
/// $Revision: 1.1 $
// ************************************************************************* //
#ifndef __HEXANION_HH__
#define __HEXANION_HH__

// ************************************************************************* //
#include "gfxtypes.h"

// ************************************************************************* //
/// DESCRIPTION: Represent the logarithm of a twist (or screw)
///  (combination of rotations and translations). This representation is 
///  ideal for blending the transformations, or perform high order
///  description of a deformation fields.
/// TODO: incorporate in gfxTypes!!!
/// TODO: extends LogTransformation
// ************************************************************************* //
class TwistLog
{
  Vec3d	_m_n, _m_cxn;
public:
  TwistLog() {}
  TwistLog(const double d): _m_n(d), _m_cxn(d) {}
  TwistLog(const double theta, const Vec3d &v, const Pnt3d &c);
  TwistLog(const Vec3d &n, const Vec3d &cxn);
  TwistLog	operator+(const TwistLog &rl) const;
  TwistLog	operator+(const Vec3d &v) const;
  TwistLog	operator-(const TwistLog &rl) const;
  Mat4d	asMatrix() const;
  Mat4d	exp() const;
  void		exp(Pnt3d &p) const;
  TwistLog	transform(const Mat4d &m) const;
  void		eqTransform(const Mat4d &m);
  TwistLog	translate(const Vec3d &t) const;
  void		operator+=(const TwistLog &l);
  void		operator-=(const TwistLog &l);
  void		operator*=(const double d);
  bool		operator!=(const TwistLog &t) const
  { return t._m_n != _m_n || t._m_cxn != _m_cxn; }
  TwistLog	operator*(const double d) const;
  friend TwistLog	operator*(const double d, const TwistLog &l)
  { return l * d; }
  TwistLog	operator/(const double d) const;
  Vec3d	operator*(const Pnt3d &p) const;
  Vec3d	operator*(const Vec3d &p) const;
  Mat4d	pow2() const;
  TwistLog	pow3() const;
  friend std::ostream	&operator<<(std::ostream &os, const TwistLog &l)
  { return os << l._m_n << " " << l._m_cxn; }
  friend std::istream	&operator>>(std::istream &is, TwistLog &l)
  { return is >> l._m_n >> l._m_cxn; }
  const Vec3d	&cxn() const { return _m_cxn; }
  const Vec3d	&n() const { return _m_n; }
  const double		norm2() const { return _m_n.norm2() + _m_cxn.norm2(); }
  void	eqSym(const Plane &p);
  TwistLog	symz() const;
  TwistLog	symx() const;
  double	&operator()(unsigned i);
  double	operator()(unsigned i) const;
};

// ************************************************************************* //
inline void	TwistLog::eqSym(const Plane &p)
{
  const Vec3d	n(-p.getv());
  const Vec3d	c(p.getc());
  const Vec3d	om(_m_n);
  const Vec3d	m(_m_cxn);

  *this = TwistLog(-om + 2.0 * (om * n) * n,
		   m - 2.0 * (c * n) * (n ^ om) - 2.0 * (m * n) * n);
}

// ************************************************************************* //
inline TwistLog	TwistLog::symz() const
{
  return TwistLog(Vec3d(-_m_n.x, -_m_n.y, _m_n.z),
		  Vec3d(_m_cxn.x, _m_cxn.y, -_m_cxn.z));
}

// ************************************************************************* //
inline TwistLog	TwistLog::symx() const
{
  return TwistLog(Vec3d(_m_n.x, -_m_n.y, -_m_n.z),
		  Vec3d(-_m_cxn.x, _m_cxn.y, _m_cxn.z));
}

// ************************************************************************* //
/// DESCRIPTION: Constructor.
// ************************************************************************* //
inline TwistLog::TwistLog(const Vec3d	&n,
			  const Vec3d	&cxn): _m_n(n), _m_cxn(cxn)
{
}

// ************************************************************************* //
/// DESCRIPTION: Builds the logarithmic form of a rotation of angle theta
///  around v, of center c.
/// v	- unit rotation vector
/// c	- center of rotation
// ************************************************************************* //
inline TwistLog::TwistLog(const double	theta,
			  const Vec3d	&v,
			  const Pnt3d	&c): 
  _m_n(v * theta), _m_cxn(Vec3d(c) ^ _m_n)
{
}

// ************************************************************************* //
/// DESCRIPTION: Addition
// ************************************************************************* //
inline TwistLog	TwistLog::operator+(const TwistLog &rl) const
{
  return TwistLog(_m_n + rl._m_n, _m_cxn + rl._m_cxn);
}

// ************************************************************************* //
/// DESCRIPTION: Adds vector v as a translation hexanion
// ************************************************************************* //
inline TwistLog	TwistLog::operator+(const Vec3d &v) const
{
  return TwistLog(_m_n, _m_cxn + v);
}

// ************************************************************************* //
/// DESCRIPTION: Substraction
// ************************************************************************* //
inline TwistLog	TwistLog::operator-(const TwistLog &rl) const
{
  return TwistLog(_m_n - rl._m_n, _m_cxn - rl._m_cxn);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the representation in an equivalent 4x4 matrix form.
// ************************************************************************* //
inline Mat4d	TwistLog::asMatrix() const
{
  return Mat4d(0.0, -_m_n.z, _m_n.y, _m_cxn.x,
		  _m_n.z, 0.0, -_m_n.x, _m_cxn.y,
		  -_m_n.y, _m_n.x, 0.0, _m_cxn.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
/// DESCRIPTION: Converts this instance into its exponential form.
/// Three cases: 1. pure translation,
///              2. translation/rotation
///              3. general case:
///  exp(log R) = I + logR + logR^2 (1-cos|n|)/n2 + logR^3(|n|-sin|n|)/(n n2)
// ************************************************************************* //
inline Mat4d	TwistLog::exp() const
{
  const double	n2(_m_n.norm2());

  if (n2 < 1.0e-64)
    return Translation(_m_cxn).asMatrix();

  const Vec3d	nxcxn(_m_n ^ _m_cxn);

  if (nxcxn.norm2() < 1.0e-64)
    {
      Quaternion	q(_m_n * 0.5);
      Vec3d		c;

      q.eqExp();
      if ((fabs(_m_n.y) <= fabs(_m_n.x)) && (fabs(_m_n.z) <= fabs(_m_n.x)))
	c = Vec3d(1.0, (_m_n.y - _m_cxn.z) / _m_n.x,
		     (_m_cxn.y + _m_n.z) / _m_n.x);
      else if (fabs(_m_n.z) <= fabs(_m_n.y))
	c = Vec3d((_m_cxn.z + _m_n.x) / _m_n.y, 1.0,
		     (_m_n.z - _m_cxn.x) / _m_n.y);
      else
	c = Vec3d((_m_cxn.x + _m_n.y) / _m_n.z,
		     (_m_n.x - _m_cxn.y) / _m_n.z, 1.0);
      return Rotation(q, c).asMatrix();
    }

  const double	n(sqrt(n2)), invn2(1.0 / n2);
  const double	cosn(cos(n)), sinn(sin(n)), k(sinn / n);
  const double	cl2((1.0 - cosn) * invn2), cl3((1.0 - k) * invn2);
  const double	nxy(_m_n.x * _m_n.y * cl2), nxx(_m_n.x * _m_n.x * cl2);
  const double	nyz(_m_n.y * _m_n.z * cl2), nyy(_m_n.y * _m_n.y * cl2);
  const double	nzx(_m_n.z * _m_n.x * cl2), nzz(_m_n.z * _m_n.z * cl2);
  const Vec3d	cn(_m_n * k);
  const Vec3d	ccxn(_m_n * ((_m_cxn * _m_n) * cl3) + _m_cxn * k);

  return Mat4d(nxx + cosn, nxy - cn.z, nzx + cn.y, nxcxn.x * cl2 + ccxn.x,
		  nxy + cn.z, nyy + cosn, nyz - cn.x, nxcxn.y * cl2 + ccxn.y,
		  nzx - cn.y, nyz + cn.x, nzz + cosn, nxcxn.z * cl2 + ccxn.z,
		  0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
/// DESCRIPTION: Applies the exponential of this transformation to the 
///  given point.
// ************************************************************************* //
inline void	TwistLog::exp(Pnt3d &q) const
{
  const double	theta2(_m_n.norm2());

  if (theta2 < 1.0e-64)
    {
      q += _m_cxn;
      return;
    }

  const Vec3d	nxcxn(_m_cxn ^ _m_n);

  if (nxcxn.norm2() < 1.0e-64)
    {
      const double	theta(sqrt(theta2)), itheta(1.0 / theta);
      const Vec3d	xi(_m_n ^ Vec3d(q) + _m_cxn);

      q += (_m_n ^ xi) * ((1.0 - cos(theta)) * itheta * itheta) +
	xi * (sin(theta) * itheta);
      return;
    }

  const Vec3d	p(q);
  const double	n(sqrt(theta2)), in2(1.0 / theta2);
  const double	a((1.0 - cos(n)) * in2), b(sin(n) / n);

  q += _m_cxn + (_m_n ^ ((_m_n ^ (p * a + _m_cxn * ((1 - b) * in2))) +
			 _m_cxn * a + p * b));
}

// ************************************************************************* //
/// DESCRIPTION: Places the transformation in the coordinate set specified 
///  by matrix m = (x y z c)
///                (0 0 0 1)
/// Constrain: x.x == y.y == z.z and x.y == y.z == z.x == 0
// ************************************************************************* //
inline TwistLog	TwistLog::transform(const Mat4d	&m) const
{
  const Vec3d	x(m.getx()), y(m.gety()), z(m.getz()), o(m.geto());
  const double		l(x.norm2());
  const double		ox(o * x), oy(o * y), oz(o * z);

  return TwistLog((_m_n.x * (y ^ z) + _m_n.y * (z ^ x) + _m_n.z * (x ^ y)) / l,
		  (_m_cxn.x + (_m_n.z * oy - _m_n.y * oz) / l) * x
		  + (_m_cxn.y + (_m_n.x * oz - _m_n.z * ox) / l) * y
		  + (_m_cxn.z + (_m_n.y * ox - _m_n.x * oy) / l) * z);
}

// ************************************************************************* //
/// DESCRIPTION: Places the transformation in the coordinate set specified 
///  by matrix m = (x y z c)
///                (0 0 0 1)
/// Constrain: x.x == y.y == z.z and x.y == y.z == z.x == 0
// ************************************************************************* //
inline void	TwistLog::eqTransform(const Mat4d	&m)
{
  const Vec3d	x(m.getx()), y(m.gety()), z(m.getz()), o(m.geto());
  const double		l(x.norm2());
  const double		ox(o * x), oy(o * y), oz(o * z);

  _m_cxn = (_m_cxn.x + (_m_n.z * oy - _m_n.y * oz) / l) * x
    + (_m_cxn.y + (_m_n.x * oz - _m_n.z * ox) / l) * y
    + (_m_cxn.z + (_m_n.y * ox - _m_n.x * oy) / l) * z;
  _m_n = (_m_n.x * (y ^ z) + _m_n.y * (z ^ x) + _m_n.z * (x ^ y)) / l;
}

// ************************************************************************* //
/// DESCRIPTION: Translates the transformation
// ************************************************************************* //
inline TwistLog	TwistLog::translate(const Vec3d &t) const
{
  return TwistLog(_m_n, _m_cxn + t ^ _m_n);
}

// ************************************************************************* //
inline void	TwistLog::operator+=(const TwistLog &l)
{
  _m_n += l._m_n;
  _m_cxn += l._m_cxn;
}

// ************************************************************************* //
inline void	TwistLog::operator-=(const TwistLog &l)
{
  _m_n -= l._m_n;
  _m_cxn -= l._m_cxn;
}

// ************************************************************************* //
inline void	TwistLog::operator*=(const double d)
{
  _m_n *= d;
  _m_cxn *= d;
}

// ************************************************************************* //
inline TwistLog	TwistLog::operator*(const double d) const
{
  return TwistLog(_m_n * d, _m_cxn * d);
}

// ************************************************************************* //
inline TwistLog	TwistLog::operator/(const double d) const
{
  const double	invd(1.0 / d);

  return TwistLog(_m_n * invd, _m_cxn * invd);
}

// ************************************************************************* //
/// DESCRIPTION: If this is the logarithm of a 4x4 transformation matrix 
///  constant in xyz, this function returns the vector field associated 
///  with this transformation, evaluated at position p.
// ************************************************************************* //
inline Vec3d	TwistLog::operator*(const Pnt3d &p) const
{
  return (_m_n ^ Vec3d(p)) + _m_cxn;
}


// ************************************************************************* //
/// DESCRIPTION: Twists a vector, i.e. only affected by a rotation.
// ************************************************************************* //
inline Vec3d	TwistLog::operator*(const Vec3d &p) const
{
  return _m_n ^ Vec3d(p);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the square of the logarithm. 
///  This expands to a matrix.
// ************************************************************************* //
inline Mat4d	TwistLog::pow2() const
{
  const double	n2(_m_n * _m_n);
  const double	nxy(_m_n.x * _m_n.y);
  const double	nyz(_m_n.y * _m_n.z);
  const double	nzx(_m_n.z * _m_n.x);
  const double	nxx(_m_n.x * _m_n.x);
  const double	nyy(_m_n.y * _m_n.y);
  const double	nzz(_m_n.z * _m_n.z);
  const Vec3d	nxcxn(_m_n ^ _m_cxn);

  return Mat4d(nxx - n2, nxy, nzx, nxcxn.x,
	       nxy, nyy - n2, nyz, nxcxn.y,
	       nzx, nyz, nzz - n2, nxcxn.z,
	       0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the cube of the logarithm.
// ************************************************************************* //
inline TwistLog	TwistLog::pow3() const
{
  const double	n2(-(_m_n * _m_n));

  return TwistLog(_m_n * n2, _m_n * (_m_cxn * _m_n) + _m_cxn * n2);
}

// ************************************************************************* //
inline double	&TwistLog::operator()(unsigned i)
{
  return (i < 3)? _m_n(i): _m_cxn(i - 3);
}

// ************************************************************************* //
inline double	TwistLog::operator()(unsigned i) const
{
  return (i < 3)? _m_n(i): _m_cxn(i - 3);
}

// ************************************************************************* //
#endif//__HEXANION_HH__
